/*
 * sample.c - random sampling of STS(19)
 *
 * Copyright (c) 2008, 2014 Petteri Kaski, Patric Ostergard, Olli Pottonen
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* The function random() is an X/Open System Interfaces Extension.
 * For C99 compatibility, we need the sixth issue. */
#define _XOPEN_SOURCE 600

#include <time.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "compression.h"

#define STS_PER_FILE 125000000
#define STS_COUNT    11084874829
#define FILE_COUNT   89
#define FILENAME_LEN 9 /* Length of "sts19-XX" including terminating '\0' */

static file_r *files[FILE_COUNT];
static int     files_initialized = 0;

static int                ull_cmp(const void*, const void*);
static void               init_rnd();
static unsigned long long rnd(unsigned long long range);

static void               init_files(const char *dir);

/* We need good pseudo-random unsigned long long int. When possible, we use
 * random() since it is usually much better than rand(). In either case we
 * need to combine several random int's to get a random long long int.
 */
#if defined(__WIN32__)
static void init_rnd() {
  srand((unsigned int)time(NULL));
}

static unsigned long long rnd(unsigned long long range) {
  assert(ULLONG_MAX / range > 10000);

  unsigned long long int r1 = rand(), r2 = rand(), r3 = rand(), r4 = rand(),
                         r5 = rand();
  r1 = r1 ^ (r2 << 13) ^ (r3 << 26) ^ (r4 << 39) ^ (r5 << 52);
  return r1 % range;
}
#else
static void init_rnd() {
  srandom((unsigned int)time(NULL));
}

static unsigned long long rnd(unsigned long long range) {
  assert(ULLONG_MAX / range > 10000);
  unsigned long long r1 = random(), r2 = random(), r3 = random();
  r1 = r1 ^ (r2 << 30) ^ (r3 << 60);
  return r1 % range;
}
#endif

static int ull_cmp(const void *v1, const void *v2) {
  unsigned long long i1 = *(const unsigned long long*)v1,
                     i2 = *(const unsigned long long*)v2;
  return (i1 > i2) - (i1 < i2);
}

const int * get_sts(long long idx)
{
  if(idx >= STS_COUNT || idx < 0) {
    fprintf(stderr, "get_sts(): invalid idx\n");
    abort();
  }
  if(!files_initialized) {
    fprintf(stderr, "get_sts(): files not opened\n");
    abort();
  }

  file_r *r = files[idx/STS_PER_FILE];
  idx       = idx % STS_PER_FILE;

  unsigned long setsize = get_set_size(r);
  unsigned long offset = get_offset(r);
  unsigned long offset_set_idx = offset/setsize;
  unsigned long set_idx = idx/setsize;
  if(offset > idx || offset_set_idx != set_idx) {
    open_set(r, set_idx);
    offset = set_idx * setsize;
  }

  assert(get_offset(r) == offset);
  assert(offset <= idx);
  assert(offset + setsize > idx);

  const int *res = NULL;
  for(; offset <= idx; offset++)
    res = decompress(r);

  assert(get_offset(r) == idx+1);

  return res;
}

/* Argument dir is the directory in which the files sts-XX are
 * located. */
static void init_files(const char *dir)
{
  if(files_initialized)
    return;

  int dir_len = strlen(dir);
  /* If necessary, add the trailing '/' or '\\' to the path. */
#if defined(__WIN32__)
  /* '\' instead of '/' */
  const char *sl = (dir_len > 0 && dir[dir_len-1] != '\\') ? "\\" : "";
#else
  const char *sl = (dir_len > 0 && dir[dir_len-1] != '/') ? "/" : "";
#endif

  for(int i = 0; i < FILE_COUNT; i++) {
    FILE *f;
    char filename[dir_len + FILENAME_LEN + 1];

    snprintf(filename, sizeof(filename), "%s%ssts19-%02d", dir, sl, i+1);
    if( (f = fopen(filename, "rb")) == NULL) {
      fprintf(stderr, "Unable to open file '%s'.\n", filename);
      abort();
    }
    files[i] = init_decompress(f);
    unsigned long sts_count = get_sts_count(files[i]);
    if( (i < FILE_COUNT-1 && sts_count != STS_PER_FILE) ||
	(i == FILE_COUNT-1 && sts_count != (STS_COUNT % STS_PER_FILE)) ) {
      fprintf(stderr, "Unexpected number (%lu) of STS's in file %s", sts_count,
	      filename);
      abort();
    }
  }

  files_initialized = 1;
}


int main(int argc, char *argv[])
{
  const char *dir = argv[1];
  int       count = 10;

  if(argc <= 1 || argc > 3)
      goto usage;

  if(argc > 2) {
    count = strtol(argv[2], NULL, 10);
    if(count <= 0)
      goto usage;
  }

  init_rnd();
  init_files(dir);

  /* Get random sample. */
  unsigned long long *indices = malloc(sizeof(unsigned long long[count]));
  for(int i = 0; i < count; i++)
    indices[i] = rnd(STS_COUNT);

  /* Make sure there are no duplicates. */
  int change = 1;
  while(change) {
    qsort(indices, count, sizeof(unsigned long long), ull_cmp);
    change = 0;
    for(int i = 1; i < count; i++)
      if(indices[i] == indices[i-1]) {
        indices[i] = rnd(STS_COUNT);
	change = 1;
      }
  }

  /* Get the actual STS's. */
  for(int i = 0; i < count; i++) {
    const int *sts = get_sts(indices[i]);

    for(int j = 0; j < 3*STS_B; j++)
      fputc('a' + sts[j], stdout);
    fputc('\n', stdout);
    continue;
  }

  return 0;

usage:
  fprintf(stderr, "usage: sampe path [count]\n");
  return 1; 
}
