/*
 * stsc.c - command line utility for (de)compressing Steiner Triple Systems
 *
 * Copyright (c) 2008, 2014 Petteri Kaski, Patric Ostergard, Olli Pottonen
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "compression.h"
#include <string.h>
#include <stdlib.h>
#include <assert.h>

int main(int argc, char *argv[])
{
    FILE *in, *out;

#define COMPRESS 0
#define DECOMPRESS 1
#define GET_SINGLE 2
    int mode;

    int next_idx = 1;
    if( next_idx >= argc)
        goto usage;

    if( strcmp(argv[next_idx], "-c") == 0 )
        mode=COMPRESS;
    else if( strcmp(argv[next_idx], "-u") == 0 )
        mode=DECOMPRESS;
    else if( strcmp(argv[next_idx], "-s") == 0 )
        mode=GET_SINGLE;
    else
        goto usage;
    next_idx++;

    long long start_idx = 0, end_idx = -2;

    if( mode == DECOMPRESS) {
        if( next_idx >= argc )
            goto usage;

        if(strcmp(argv[next_idx], "-f") == 0) {
	    if( ++next_idx >= argc )
                goto usage;
	    start_idx = strtoll(argv[next_idx++], NULL, 10)-1;
            if( next_idx >= argc )
                goto usage;
	}

	if(strcmp(argv[next_idx], "-t") == 0) {
	    if( ++next_idx >= argc )
                goto usage;
	    end_idx = strtoll(argv[next_idx++], NULL, 10)-1;
	}
    }	  

    if( next_idx >= argc)
        goto usage;
    if( strcmp(argv[next_idx], "-") == 0 )
        in = stdin;
    else if( (in = fopen(argv[next_idx], "rb")) == NULL ) {
        fprintf(stderr, "Unable to open file %s\n", argv[2]);
        return 1;
    }
    next_idx++;

    if( mode == COMPRESS || mode == DECOMPRESS ) {
        if( next_idx >= argc)
            goto usage;

        if( strcmp(argv[next_idx], "-") == 0 )
            out = stdout;
        else if( (out = fopen(argv[next_idx], "wb")) == NULL ) {
            fprintf(stderr, "Unable to open file %s\n", argv[3]);
            return 1;
        }
	next_idx++;
    } else // mode == GET_SINGLE
        out = stdout;
    
    if( next_idx != argc)
        /* All arguments should be read now. */
        goto usage;

    /* output buffer */
    char csts[3*STS_B+2];
    csts[3*STS_B] = '\n';
    csts[3*STS_B+1] = '\0';

    if( mode == COMPRESS ) {
        /* input buffer */
        int t[3*STS_B];
        file_w *w;
        w = init_compress(out, 0);

        for(;;) {
            /* Read STS */
            for(int i=0 ; i<3*STS_B ; ++i)
                t[i] = fgetc(in) - 'a';
            fgetc(in); // Ignore newline
            if( feof(in) )
                break;
            /* Compress & output the STS */
            compress(t, w);
        }

        free_compress(w);
        return 0;

    } else if( mode == DECOMPRESS ) {
        file_r *r;
        r = init_decompress(in);

        unsigned stscount = get_sts_count(r), setsize = get_set_size(r);
        fprintf(stderr, "Setsize: %u, number of systems: %u\n",
		setsize, stscount);

	if(end_idx == -2)
            end_idx = stscount-1;
        if(end_idx < 0 || end_idx < start_idx) {
            fprintf(stderr, "Invalid end index %lld\n", end_idx);
            abort();
	}
	if(start_idx < 0 || start_idx >= stscount) {
            fprintf(stderr, "Invalid start index %lld\n", end_idx);
            abort();
	}

	open_set(r, start_idx/setsize);
	for(int i = setsize*(start_idx/setsize); i < start_idx; i++)
            decompress(r);

	assert(get_offset(r) == start_idx);

        const int *sts;
        for(int i=start_idx ; i<=end_idx ; ++i) {
            /* Decompress */
            sts = decompress(r);
            for(int j=0 ; j<3*STS_B ; ++j)
                csts[j] = sts[j] + 'a';
            fputs(csts, out);
        }
	assert(get_offset(r) == end_idx+1);

        free_decompress(r);
        return 0;

    } else { // mode == GET_SINGLE
        file_r *r;
        r = init_decompress(in);

        long long idx = strtoll(argv[3], NULL, 10)-1;
        if( idx < 0 || idx >= get_sts_count(r) ) {
            fprintf(stderr, "Invalid parameter '%s'\n.", argv[3]);
                return 1;
        }

        const int *sts = NULL;
        int setsize = get_set_size(r);
        open_set(r, idx/setsize);
        for(int i=0 ; i <= (int)(idx % setsize) ; ++i)
            sts = decompress(r);

        for(int i=0 ; i<3*STS_B ; ++i)
            csts[i] = sts[i] + 'a';
        fputs(csts, stdout);

        free_decompress(r);
        return 0;
    }

  usage:
    fprintf(stderr, "usage:\n%s -u [-f startindex] [-t endindex] <inputfile> <outputfile> (decompress) or\n"
	    "stsc -c <inputfile> <outputfile> (compress) or\n"
	    "stsc -s <inputfile>\n", argv[0]);
    return 1;
}
